@@ -15,6 +15,10 @@
 """
 from __future__ import print_function
 
+try:
+    from collections.abc import MutableMapping
+except ImportError:
+    from collections import MutableMapping
 from contextlib import contextmanager
 import doctest
 import os
@@ -64,8 +68,9 @@ class OutputFormatter(object):
     progress = property(lambda self: self.options.progress)
     verbose = property(lambda self: self.options.verbose)
     in_subprocess = property(
-        lambda self: self.options.resume_layer is not None and
-                     self.options.processes > 1)
+        lambda self: (
+            self.options.resume_layer is not None and
+            self.options.processes > 1))
 
     def compute_max_width(self):
         """Try to determine the terminal width."""
@@ -169,10 +174,10 @@ class OutputFormatter(object):
     def summary(self, n_tests, n_failures, n_errors, n_seconds,
                 n_skipped=0):
         """Summarize the results of a single test layer."""
-        print ("  Ran %s tests with %s failures, %s errors and "
-               "%s skipped in %s."
-               % (n_tests, n_failures, n_errors, n_skipped,
-                   self.format_seconds(n_seconds)))
+        print("  Ran %s tests with %s failures, %s errors and "
+              "%s skipped in %s."
+              % (n_tests, n_failures, n_errors, n_skipped,
+                  self.format_seconds(n_seconds)))
 
     def totals(self, n_tests, n_failures, n_errors, n_seconds,
                n_skipped=0):
@@ -323,7 +328,7 @@ class OutputFormatter(object):
         sys.stdout.write(s)
         self.test_width += len(s) + 1
 
-    def test_error(self, test, seconds, exc_info):
+    def test_error(self, test, seconds, exc_info, stdout=None, stderr=None):
         """Report that an error occurred while running a test.
 
         Should be called right after start_test().
@@ -334,9 +339,10 @@ class OutputFormatter(object):
             print(" (%s)" % self.format_seconds_short(seconds))
         print()
         self.print_traceback("Error in test %s" % test, exc_info)
+        self.print_std_streams(stdout, stderr)
         self.test_width = self.last_width = 0
 
-    def test_failure(self, test, seconds, exc_info):
+    def test_failure(self, test, seconds, exc_info, stdout=None, stderr=None):
         """Report that a test failed.
 
         Should be called right after start_test().
@@ -347,6 +353,7 @@ class OutputFormatter(object):
             print(" (%s)" % self.format_seconds_short(seconds))
         print()
         self.print_traceback("Failure in test %s" % test, exc_info)
+        self.print_std_streams(stdout, stderr)
         self.test_width = self.last_width = 0
 
     def print_traceback(self, msg, exc_info):
@@ -355,6 +362,21 @@ class OutputFormatter(object):
         print(msg)
         print(self.format_traceback(exc_info))
 
+    def print_std_streams(self, stdout, stderr):
+        """Emit contents of buffered standard streams."""
+        if stdout:
+            sys.stdout.write("Stdout:\n")
+            sys.stdout.write(stdout)
+            if not stdout.endswith("\n"):
+                sys.stdout.write("\n")
+            sys.stdout.write("\n")
+        if stderr:
+            sys.stderr.write("Stderr:\n")
+            sys.stderr.write(stderr)
+            if not stderr.endswith("\n"):
+                sys.stderr.write("\n")
+            sys.stderr.write("\n")
+
     def format_traceback(self, exc_info):
         """Format the traceback."""
         v = exc_info[1]
@@ -476,12 +498,14 @@ class ColorfulOutputFormatter(OutputFormatter):
                   '?': 'character-diffs',
                   '@': 'diff-chunk',
                   '*': 'diff-chunk',
-                  '!': 'actual-output',}
+                  '!': 'actual-output',
+                  }
 
     prefixes = [('dark', '0;'),
                 ('light', '1;'),
                 ('bright', '1;'),
-                ('bold', '1;'),]
+                ('bold', '1;'),
+                ]
 
     colorcodes = {'default': 0, 'normal': 0,
                   'black': 30,
@@ -493,10 +517,10 @@ class ColorfulOutputFormatter(OutputFormatter):
                   'cyan': 36,
                   'grey': 37, 'gray': 37, 'white': 37}
 
-    slow_test_threshold = 10.0 # seconds
+    slow_test_threshold = 10.0  # seconds
 
     def color_code(self, color):
-        """Convert a color description (e.g. 'lightgray') to a terminal code."""
+        """Convert a color description (e.g. 'lightred') to a terminal code."""
         prefix_code = ''
         for prefix, code in self.prefixes:
             if color.startswith(prefix):
@@ -669,7 +693,8 @@ class ColorfulOutputFormatter(OutputFormatter):
                     print(line)
             elif line.startswith('    ') or line.strip() == '':
                 if colorize_diff and len(line) > 4:
-                    color = self.diff_color.get(line[4], color_of_indented_text)
+                    color = self.diff_color.get(
+                        line[4], color_of_indented_text)
                     print(self.colorize(color, line))
                 else:
                     if line.strip() != '':
@@ -792,6 +817,37 @@ class _RunnableDecorator(object):
         self.decorated.status(**kwargs)
 
 
+class _SortedDict(MutableMapping, object):
+    """A dict that always returns items in sorted order.
+
+    This differs from collections.OrderedDict in that it returns items in
+    *sorted* order, not in insertion order.
+
+    We use this as a workaround for the fact that
+    testtools.ExtendedToStreamDecorator doesn't sort the details dict when
+    encoding it, which makes it difficult to write stable doctests for
+    subunit v2 output.
+    """
+
+    def __init__(self, items):
+        self._dict = dict(items)
+
+    def __getitem__(self, key):
+        return self._dict[key]
+
+    def __setitem__(self, key, value):
+        self._dict[key] = value
+
+    def __delitem__(self, key):
+        del self._dict[key]
+
+    def __iter__(self):
+        return iter(sorted(self._dict))
+
+    def __len__(self):
+        return len(self._dict)
+
+
 class SubunitOutputFormatter(object):
     """A subunit output formatter.
 
@@ -885,6 +941,15 @@ class SubunitOutputFormatter(object):
             self._subunit.addError(test, exc_info)
             self._subunit.stopTest(test)
 
+    def _emit_failure(self, failure_id, tag, exc_info):
+        """Emit an failure to the subunit stream.
+
+        Use this to pass on information about failures that occur outside of
+        tests.
+        """
+        test = FakeTest(failure_id)
+        self._subunit.addFailure(test, exc_info)
+
     def _enter_layer(self, layer_name):
         """Tell subunit that we are entering a layer."""
         self._subunit.tags(['zope:layer:%s' % (layer_name,)], [])
@@ -1043,20 +1108,20 @@ class SubunitOutputFormatter(object):
 
     def refcounts(self, rc, prev):
         """Report a change in reference counts."""
-        details = {
+        details = _SortedDict({
             'sys-refcounts': text_content(str(rc)),
             'changes': text_content(str(rc - prev)),
-        }
+        })
         # XXX: Emit the details dict as JSON?
         self._emit_fake_test(self.TAG_REFCOUNTS, self.TAG_REFCOUNTS, details)
 
     def detailed_refcounts(self, track, rc, prev):
         """Report a change in reference counts, with extra detail."""
-        details = {
+        details = _SortedDict({
             'sys-refcounts': text_content(str(rc)),
             'changes': text_content(str(rc - prev)),
             'track': text_content(str(track.delta)),
-        }
+        })
         self._emit_fake_test(self.TAG_REFCOUNTS, self.TAG_REFCOUNTS, details)
 
     def start_set_up(self, layer_name):
@@ -1089,6 +1154,11 @@ class SubunitOutputFormatter(object):
             self._subunit.stopTest(test)
         self._enter_layer(layer_name)
 
+    def layer_failure(self, failure_type, exc_info):
+        layer_name, start_time = self._last_layer
+        self._emit_failure(
+            '%s:%s' % (layer_name, failure_type), self.TAG_LAYER, exc_info)
+
     def start_tear_down(self, layer_name):
         """Report that we're tearing down a layer.
 
@@ -1180,13 +1250,26 @@ class SubunitOutputFormatter(object):
         else:
             unicode_tb = traceback
 
-        return {
+        return _SortedDict({
             'traceback': Content(
                 self.TRACEBACK_CONTENT_TYPE,
                 lambda: [unicode_tb.encode('utf8')]),
-        }
+        })
 
-    def test_error(self, test, seconds, exc_info):
+    def _add_std_streams_to_details(self, details, stdout, stderr):
+        """Add buffered standard stream contents to a subunit details dict."""
+        if stdout:
+            if isinstance(stdout, bytes):
+                stdout = stdout.decode('utf-8', 'replace')
+            details['test-stdout'] = Content(
+                self.PLAIN_TEXT, lambda: [stdout.encode('utf-8')])
+        if stderr:
+            if isinstance(stderr, bytes):
+                stderr = stderr.decode('utf-8', 'replace')
+            details['test-stderr'] = Content(
+                self.PLAIN_TEXT, lambda: [stderr.encode('utf-8')])
+
+    def test_error(self, test, seconds, exc_info, stdout=None, stderr=None):
         """Report that an error occurred while running a test.
 
         Should be called right after start_test().
@@ -1195,9 +1278,10 @@ class SubunitOutputFormatter(object):
         """
         self._emit_timestamp()
         details = self._exc_info_to_details(exc_info)
+        self._add_std_streams_to_details(details, stdout, stderr)
         self._subunit.addError(test, details=details)
 
-    def test_failure(self, test, seconds, exc_info):
+    def test_failure(self, test, seconds, exc_info, stdout=None, stderr=None):
         """Report that a test failed.
 
         Should be called right after start_test().
@@ -1206,6 +1290,7 @@ class SubunitOutputFormatter(object):
         """
         self._emit_timestamp()
         details = self._exc_info_to_details(exc_info)
+        self._add_std_streams_to_details(details, stdout, stderr)
         self._subunit.addFailure(test, details=details)
 
     def stop_test(self, test):
