@@ -55,10 +55,11 @@ import zope.testrunner.tb_format
 import zope.testrunner.shuffle
 
 try:
-    import Queue # Python 2
+    import Queue  # Python 2
 except ImportError:
     # Python 3
-    import queue as Queue # Python 3
+    import queue as Queue  # Python 3
+
 
 class UnexpectedSuccess(Exception):
     pass
@@ -94,11 +95,12 @@ class Runner(object):
 
     .. versionchanged:: 4.8.0
        Add the *warnings* keyword argument. If this is ``None`` (the default)
-       and the user hasn't configured Python otherwise with command-line arguments
-       or environment variables, we will enable the default warnings, including
-       ``DeprecationWarning``, when running tests. Otherwise, it can be any
-       string acceptable to :func:`warnings.simplefilter` and that filter will
-       be in effect while running tests.
+       and the user hasn't configured Python otherwise with command-line
+       arguments or environment variables, we will enable the default
+       warnings, including ``DeprecationWarning``, when running tests.
+       Otherwise, it can be any string acceptable to
+       :func:`warnings.simplefilter` and that filter will be in effect while
+       running tests.
 
     """
 
@@ -180,9 +182,9 @@ class Runner(object):
 
             # Late setup
             #
-            # Some system tools like profilers are really bad with stack frames.
-            # E.g. hotshot doesn't like it when we leave the stack frame that we
-            # called start() from.
+            # Some system tools like profilers are really bad with stack
+            # frames.  E.g. hotshot doesn't like it when we leave the stack
+            # frame that we called start() from.
             for feature in self.features:
                 feature.late_setup()
 
@@ -268,9 +270,10 @@ class Runner(object):
                 # noisy.  The -Wd and -Wa flags can be used to bypass this
                 # only when self.warnings is None.
                 if self.warnings in ['default', 'always']:
-                    warnings.filterwarnings('module',
-                                            category=DeprecationWarning,
-                                            message=r'Please use assert\w+ instead.')
+                    warnings.filterwarnings(
+                        'module',
+                        category=DeprecationWarning,
+                        message=r'Please use assert\w+ instead.')
             yield
 
     def run_tests(self):
@@ -317,11 +320,22 @@ class Runner(object):
         if setup_layers:
             if self.options.resume_layer is None:
                 self.options.output.info("Tearing down left over layers:")
-            tear_down_unneeded(self.options, (), setup_layers, True)
+            tear_down_unneeded(
+                self.options, (), setup_layers, self.errors, optional=True)
 
         self.failed = bool(self.import_errors or self.failures or self.errors)
 
 
+def handle_layer_failure(failure_type, output, errors):
+    if hasattr(output, 'layer_failure'):
+        output.layer_failure(failure_type.subunit_label, sys.exc_info())
+    else:
+        f = StringIO()
+        traceback.print_exc(file=f)
+        output.error(f.getvalue())
+    errors.append((failure_type, sys.exc_info()))
+
+
 def run_tests(options, tests, name, failures, errors, skipped, import_errors):
     repeat = options.repeat or 1
     repeat_range = iter(range(repeat))
@@ -364,7 +378,7 @@ def run_tests(options, tests, name, failures, errors, skipped, import_errors):
                         test.debug()
                     except KeyboardInterrupt:
                         raise
-                    except:
+                    except BaseException:
                         result.addError(
                             test,
                             sys.exc_info()[:2] + (sys.exc_info()[2].tb_next, ),
@@ -389,14 +403,16 @@ def run_tests(options, tests, name, failures, errors, skipped, import_errors):
         t = time.time() - t
         output.stop_tests()
         failures.extend(result.failures)
+        n_failures = len(result.failures)
         if hasattr(result, 'unexpectedSuccesses'):
             # Python versions prior to 2.7 do not have the concept of
             # unexpectedSuccesses.
             failures.extend(result.unexpectedSuccesses)
+            n_failures += len(result.unexpectedSuccesses)
         skipped.extend(result.skipped)
         errors.extend(result.errors)
         output.summary(n_tests=result.testsRun,
-                       n_failures=len(failures),
+                       n_failures=n_failures,
                        n_errors=len(result.errors) + len(import_errors),
                        n_seconds=t,
                        n_skipped=len(result.skipped))
@@ -439,10 +455,10 @@ def run_layer(options, layer_name, layer, tests, setup_layers,
     output = options.output
     gathered = []
     gather_layers(layer, gathered)
-    needed = dict([(l, 1) for l in gathered])
+    needed = dict([(ly, 1) for ly in gathered])
     if options.resume_number != 0:
         output.info("Running %s tests:" % layer_name)
-    tear_down_unneeded(options, needed, setup_layers)
+    tear_down_unneeded(options, needed, setup_layers, errors)
 
     if options.resume_layer is not None:
         output.info_suboptimal("  Running in a subprocess.")
@@ -451,11 +467,10 @@ def run_layer(options, layer_name, layer, tests, setup_layers,
         setup_layer(options, layer, setup_layers)
     except zope.testrunner.interfaces.EndRun:
         raise
+    except MemoryError:
+        raise
     except Exception:
-        f = StringIO()
-        traceback.print_exc(file=f)
-        output.error(f.getvalue())
-        errors.append((SetUpLayerFailure(layer), sys.exc_info()))
+        handle_layer_failure(SetUpLayerFailure(layer), output, errors)
         return 0
     else:
         return run_tests(options, tests, layer_name, failures, errors, skipped,
@@ -464,6 +479,8 @@ def run_layer(options, layer_name, layer, tests, setup_layers,
 
 class SetUpLayerFailure(unittest.TestCase):
 
+    subunit_label = 'setUp'
+
     def __init__(self, layer):
         super(SetUpLayerFailure, self).__init__()
         self.layer = layer
@@ -472,9 +489,23 @@ class SetUpLayerFailure(unittest.TestCase):
         pass
 
     def __str__(self):
-        return "Layer: %s" % (name_from_layer(self.layer))
+        return "Layer: %s.setUp" % (name_from_layer(self.layer))
 
 
+class TearDownLayerFailure(unittest.TestCase):
+
+    subunit_label = 'tearDown'
+
+    def __init__(self, layer):
+        super(TearDownLayerFailure, self).__init__()
+        self.layer = layer
+
+    def runTest(self):
+        pass
+
+    def __str__(self):
+        return "Layer: %s.tearDown" % (name_from_layer(self.layer))
+
 
 def spawn_layer_in_subprocess(result, script_parts, options, features,
                               layer_name, layer, failures, errors, skipped,
@@ -495,7 +526,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
 
         # this is because of a bug in Python (http://www.python.org/sf/900092)
         if (options.profile == 'hotshot'
-            and sys.version_info[:3] <= (2, 4, 1)):
+                and sys.version_info[:3] <= (2, 4, 1)):
             args.insert(1, '-O')
 
         debugargs = args  # save them before messing up for windows
@@ -507,10 +538,22 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
         for feature in features:
             feature.layer_setup(layer)
 
-        child = subprocess.Popen(args, shell=False, stdin=subprocess.PIPE,
+        child = subprocess.Popen(
+            args, shell=False, stdin=subprocess.PIPE,
             stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd,
             close_fds=not sys.platform.startswith('win'))
 
+        def reader_thread(f, buf):
+            buf.append(f.read())
+
+        # Start reading stderr in a thread.  This means we don't hang if the
+        # subprocess writes more to stderr than the pipe capacity.
+        stderr_buf = []
+        stderr_thread = threading.Thread(
+            target=reader_thread, args=(child.stderr, stderr_buf))
+        stderr_thread.daemon = True
+        stderr_thread.start()
+
         while True:
             try:
                 while True:
@@ -520,10 +563,10 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                     # return its lines as a batch). We don't want too much
                     # buffering because this foils automatic and human monitors
                     # trying to verify that the subprocess is still alive.
-                    l = child.stdout.readline()
-                    if not l:
+                    line = child.stdout.readline()
+                    if not line:
                         break
-                    result.write(l)
+                    result.write(line)
             except IOError as e:
                 if e.errno == errno.EINTR:
                     # If the subprocess dies before we finish reading its
@@ -536,8 +579,9 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             else:
                 break
 
-        # Now stderr should be ready to read the whole thing.
-        errlines = child.stderr.read().splitlines()
+        # Now we should be able to finish reading stderr.
+        stderr_thread.join()
+        errlines = stderr_buf[0].splitlines()
         erriter = iter(errlines)
         nfail = nerr = 0
         for line in erriter:
@@ -553,7 +597,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
             if options.verbose >= 1:
                 errmsg += "\nChild command line: %s" % debugargs
             if (options.verbose >= 2 or
-                (options.verbose == 1 and len(errlines) < 20)):
+                    (options.verbose == 1 and len(errlines) < 20)):
                 errmsg += ("\nChild stderr was:\n" +
                            "\n".join("  " + line.decode('utf-8', 'replace')
                                      for line in errlines))
@@ -566,7 +610,6 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                                      for line in errlines[-10:]))
             output.error_with_banner(errmsg)
 
-
         while nfail > 0:
             nfail -= 1
             # Doing erriter.next().strip() confuses the 2to3 fixer, so
@@ -649,7 +692,9 @@ class ImmediateSubprocessResult(AbstractSubprocessResult):
         self.stream.flush()
 
 
-_is_dots = re.compile(br'\.+(\r\n?|\n)').match # Windows sneaks in a \r\n.
+_is_dots = re.compile(br'\.+(\r\n?|\n)').match  # Windows sneaks in a \r\n.
+
+
 class KeepaliveSubprocessResult(AbstractSubprocessResult):
     "Keeps stdout for later processing; sends marks to queue to show activity."
 
@@ -742,34 +787,39 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
 
         # Help keep-alive monitors (human or automated) keep up-to-date.
         stdout.flush()
-        time.sleep(0.01) # Keep the loop from being too tight.
+        time.sleep(0.01)  # Keep the loop from being too tight.
 
     # Return the total number of tests run.
     return sum(r.num_ran for r in results)
 
 
-def tear_down_unneeded(options, needed, setup_layers, optional=False):
+def tear_down_unneeded(options, needed, setup_layers, errors, optional=False):
     # Tear down any layers not needed for these tests. The unneeded layers
     # might interfere.
-    unneeded = [l for l in setup_layers if l not in needed]
+    unneeded = [layer for layer in setup_layers if layer not in needed]
     unneeded = order_by_bases(unneeded)
     unneeded.reverse()
     output = options.output
-    for l in unneeded:
-        output.start_tear_down(name_from_layer(l))
+    for layer in unneeded:
+        output.start_tear_down(name_from_layer(layer))
         t = time.time()
         try:
             try:
-                if hasattr(l, 'tearDown'):
-                    l.tearDown()
+                if hasattr(layer, 'tearDown'):
+                    layer.tearDown()
             except NotImplementedError:
                 output.tear_down_not_supported()
                 if not optional:
-                    raise CanNotTearDown(l)
+                    raise CanNotTearDown(layer)
+            except MemoryError:
+                raise
+            except Exception:
+                handle_layer_failure(
+                    TearDownLayerFailure(layer), output, errors)
             else:
                 output.stop_tear_down(time.time() - t)
         finally:
-            del setup_layers[l]
+            del setup_layers[layer]
 
 
 cant_pm_in_subprocess_message = """
@@ -790,6 +840,8 @@ def setup_layer(options, layer, setup_layers):
         if hasattr(layer, 'setUp'):
             try:
                 layer.setUp()
+            except MemoryError:
+                raise
             except Exception:
                 if options.post_mortem:
                     if options.resume_layer:
@@ -821,6 +873,10 @@ class TestResult(unittest.TestResult):
         for test in tests:
             count += test.countTestCases()
         self.count = count
+        self._stdout_buffer = None
+        self._stderr_buffer = None
+        self._original_stdout = sys.stdout
+        self._original_stderr = sys.stderr
 
     def testSetUp(self):
         """A layer may define a setup method to be called before each
@@ -842,10 +898,52 @@ class TestResult(unittest.TestResult):
             if hasattr(layer, 'testTearDown'):
                 layer.testTearDown()
 
+    def _makeBufferedStdStream(self):
+        """Make a buffered stream to replace a standard stream."""
+        if sys.version_info[0] >= 3:
+            # The returned stream needs to have a 'buffer' attribute, since
+            # some tests may expect that to exist on standard streams, and a
+            # 'getvalue' method for the convenience of _restoreStdStreams.
+            # This requires some care.
+            class BufferedStandardStream(io.TextIOWrapper):
+                def getvalue(self):
+                    return self.buffer.getvalue().decode(
+                        encoding=self.encoding, errors=self.errors)
+
+            return BufferedStandardStream(
+                io.BytesIO(), newline='\n', write_through=True)
+        else:
+            return StringIO()
+
+    def _setUpStdStreams(self):
+        """Set up buffered standard streams, if requested."""
+        if self.options.buffer:
+            if self._stdout_buffer is None:
+                self._stdout_buffer = self._makeBufferedStdStream()
+            if self._stderr_buffer is None:
+                self._stderr_buffer = self._makeBufferedStdStream()
+            sys.stdout = self._stdout_buffer
+            sys.stderr = self._stderr_buffer
+
+    def _restoreStdStreams(self):
+        """Restore the buffered standard streams and return any contents."""
+        if self.options.buffer:
+            stdout = sys.stdout.getvalue()
+            stderr = sys.stderr.getvalue()
+            sys.stdout = self._original_stdout
+            sys.stderr = self._original_stderr
+            self._stdout_buffer.seek(0)
+            self._stdout_buffer.truncate(0)
+            self._stderr_buffer.seek(0)
+            self._stderr_buffer.truncate(0)
+            return stdout, stderr
+        else:
+            return None, None
+
     def startTest(self, test):
         self.testSetUp()
         unittest.TestResult.startTest(self, test)
-        testsRun = self.testsRun - 1 # subtract the one the base class added
+        testsRun = self.testsRun - 1  # subtract the one the base class added
         count = test.countTestCases()
         self.testsRun = testsRun + count
 
@@ -854,17 +952,23 @@ class TestResult(unittest.TestResult):
         self._threads = threading.enumerate()
         self._start_time = time.time()
 
+        self._setUpStdStreams()
+
     def addSuccess(self, test):
+        self._restoreStdStreams()
         t = max(time.time() - self._start_time, 0.0)
         self.options.output.test_success(test, t)
 
     def addSkip(self, test, reason):
+        self._restoreStdStreams()
         unittest.TestResult.addSkip(self, test, reason)
         self.options.output.test_skipped(test, reason)
 
     def addError(self, test, exc_info):
+        stdout, stderr = self._restoreStdStreams()
         self.options.output.test_error(test, time.time() - self._start_time,
-                                       exc_info)
+                                       exc_info,
+                                       stdout=stdout, stderr=stderr)
 
         unittest.TestResult.addError(self, test, exc_info)
 
@@ -879,8 +983,10 @@ class TestResult(unittest.TestResult):
             self.stop()
 
     def addFailure(self, test, exc_info):
+        stdout, stderr = self._restoreStdStreams()
         self.options.output.test_failure(test, time.time() - self._start_time,
-                                         exc_info)
+                                         exc_info,
+                                         stdout=stdout, stderr=stderr)
 
         unittest.TestResult.addFailure(self, test, exc_info)
 
@@ -892,15 +998,18 @@ class TestResult(unittest.TestResult):
             self.stop()
 
     def addExpectedFailure(self, test, exc_info):
+        self._restoreStdStreams()
         t = max(time.time() - self._start_time, 0.0)
         self.options.output.test_success(test, t)
 
         unittest.TestResult.addExpectedFailure(self, test, exc_info)
 
     def addUnexpectedSuccess(self, test):
+        stdout, stderr = self._restoreStdStreams()
         self.options.output.test_error(
             test, time.time() - self._start_time,
-            (UnexpectedSuccess, UnexpectedSuccess(), None))
+            (UnexpectedSuccess, UnexpectedSuccess(), None),
+            stdout=stdout, stderr=stderr)
 
         unittest.TestResult.addUnexpectedSuccess(self, test)
 
@@ -911,7 +1020,8 @@ class TestResult(unittest.TestResult):
                                                       " as a subprocess!")
             else:
                 # XXX: what exc_info? there's no exc_info!
-                zope.testrunner.debug.post_mortem(exc_info)
+                # flake8 is correct, but keep it quiet for now ...
+                zope.testrunner.debug.post_mortem(exc_info)  # noqa: F821
         elif self.options.stop_on_error:
             self.stop()
 
@@ -987,7 +1097,7 @@ def layer_sort_key(layer):
         key.append(layer)
 
     _gather(layer)
-    return tuple(name_from_layer(l) for l in key if l != UnitTests)
+    return tuple(name_from_layer(ly) for ly in key if ly != UnitTests)
 
 
 def order_by_bases(layers):
@@ -1023,8 +1133,8 @@ class FakeInputContinueGenerator:
     def readline(self):
         print('c\n')
         print('*'*70)
-        print ("Can't use pdb.set_trace when running a layer"
-               " as a subprocess!")
+        print("Can't use pdb.set_trace when running a layer"
+              " as a subprocess!")
         print('*'*70)
         print()
         return 'c\n'
